// Copyright(c) 2oo1 by Mental Vortex
#ifndef _L3D_H_
#define _L3D_H_
// 
// Description:
// 
// Lionhead 3D File Format .l3d
// 

//
// Update by Darkyoda
// Found out exact use of LH3D_BaseHeader::mSkinOffset 
//		In a G3D File it points to the next object, maybe a G3D, maybe something else
// 
// 
// 




// Includes
#include			"math.h"						// This include file includes the math functions




//
// This is the Base Header structure
//
struct LH3D_BaseHeader
{
	DWORD				mID;					// 'L3D0'

	short				mUnk_1;
	short				mUnk_2;

/*
//
// Addition by Scorpio9a
// email : scorpio9a@ra2maincenter.tmfweb.nl
//
// Type ID (I think mUnk_1 & mUnk_2 is 1 DWORD, not sure, but it looks like it.)

 Animal Types:
 **** Bear Types		****
 Bear			= HEX: 8321, Chars: '8!'
 Polar Bear		= HEX: 8321, Chars: '8!'

 **** Human Types		****
 Greek			= HEX: 3F21, Chars: '?!'

 **** Horse Types		****
 Horse			= HEX: 3021, Chars: '0!'
 Cow			= HEX: 3021, Chars: '0!'
 Rhino			= HEX: 3021, Chars: '0!' // What is a Rhino doing in here?
 Sheep			= HEX: 3021, Chars: '0!'
 Zebra			= HEX: 3021, Chars: '0!'

 **** Cat Types		****
 Tiger			= HEX: 3321, Chars: '3!'
 Leopard		= HEX: 3321, Chars: '3!'
 Lion			= HEX: 3321, Chars: '3!'
 Wolf			= HEX: 3321, Chars: '3!'

 **** Monkey Types	****
 Ape			= HEX: 4C21, Chars: 'L!'
 Chimp			= HEX: 4C21, Chars: 'L!'
 Gorilla		= HEX: 4C21, Chars: 'L!'
 Mandrill		= HEX: 4C21, Chars: 'L!'

 **** Turtel Types	****
 Tortoise		= HEX: 3C21, Chars: '<!'

 Other Types:
 **** Hand Types		****
 Hand			= HEX: 1623, Chars:	' #'

 **** Static Mapobjs  ****
 Waterfall		= HEX: 0020
 Arche			= HEX: 0020
 Dinosaur		= HEX: 00A0

 **** Dyn. Mapobjects ****
 Coffre			= HEX: 0A23
 Chest			= HEX: 0022


// DarkYoda in Addition :
// I think its a Bitarray of Flags.
// for example, all creatures have bit 0 and 5 of the second byte set to 1
// the hand has bits 0,1,5 set to 1

*/

	DWORD				mSkinOffset;			// Skin Offset from LH3D_BaseHeader (+4 to Skindata)

	DWORD				mnMeshes;				// Count of Meshes		
	DWORD				mMeshListOffset;		// Offset from LH3D_BaseHeader to an array of pointers to Meshes

	DWORD				mUnk_3;
	DWORD				mUnk_4;	
	DWORD				mUnk_5;	
	DWORD				mUnk_6;	
	DWORD				mUnk_7;	
	DWORD				mUnk_8;	
	DWORD				mUnk_9;	
	DWORD				mUnk_10;	
	DWORD				mUnk_11;	

	DWORD				mnSkins;				// Number of Skins
	DWORD				mSkinListOffset;	

	DWORD				mUnk_12;	
	DWORD				mUnk_13;	
	DWORD				mpSkinName;				
	//
	// Thesis about mpSkinName
	//
	// DWORD		Size of SkinName Block
	// DWORD		Number of SkinNames
	// char			skinname[x]
	//
	//
};



//
// This is the Mesh structure
//
struct LH3D_Mesh
{
	// FLAGS :
	byte				mUnk_1;					
	byte				mUnk_2;					// 20 : nodraw 
	byte				mUnk_3;
	byte				mUnk_4;					// e0 : transparent
												

	DWORD				mnSubMeshes;			// Number of SubMeshes in this Mesh
	DWORD				mSubMeshListOffset;		// Offset from LH3D_BaseHeader to an array of offsets to SubMeshes

	DWORD				mnBones;				// Number of Bones in this Mesh
	DWORD				mpBones;				// Offset from LH3D_BaseHeader to LH3D_Bones
};



//
// This is the SubMesh structure
//
struct LH3D_SubMesh
{
	DWORD				mUnk_1;
	DWORD				mUnk_2;
	DWORD				mSkinID;				// Skin ID
	DWORD				mUnk_4;

	DWORD				mnVertices;				// Number of Vertices in this Submesh
	DWORD				mpVertices;				// Offset from LH3D_BaseHeader to an array of LH3D_Vertex structures

	DWORD				mnTriangles;			// Number of Triangles in this Submesh
	DWORD				mpTriangles;			// Offset from LH3D_BaseHeader to an array of LH3D_Triangle structures

//
// I got an idea about this :
// 2 shorts
// first short gives the number of vertices for this bone, RLE means that it's always increasing
// second short gives the index of the bone
//
	DWORD				mszBoneVertLUT;			// Size of the Bone-Vertex	LUT
	DWORD				mpBoneVertLUT;			// Offset from LH3D_BaseHeader to the LUT

//
// Thanx Chris Cookson (cjcookson@hotmail.com)
//
	DWORD				mnVerrtexBlends;		// Number of Vertices lying at bone seams, so they got to be blended
	DWORD				mpVertexBlends;			// offset
};



//
// Vertices @ seams are blended after applying bones, to ensure no holes pop up
//
// Thanx Chris Cookson (cjcookson@hotmail.com)
//
struct LH3D_VertexBlend
{
	short				mIndex1;
	short				mIndex2;
	float				mWeight;
};


//
// This is the Vertex structure
//
struct LH3D_Vertex
{
	Vec3				mPosition;				// Position (bodyspace)
	Vec2				mTexCoords;				// Texture Coordinates
	Vec3				mNormal;				// Normal (bodyspace)
};



//
// This is the Triangle structure
//
struct LH3D_Triangle
{
	short				mIndices[3];						// The Indicies 
};



//
// This is the Bone RLE LUT
//
struct LH3D_BoneVert
{
	short				mnVertices;
	short				mBoneIndex;
};



//
// This is the Bone structure
//
struct LH3D_Bone
{
	//
	// Hum this is getting complex.... but i got an idea:
	// unk_0 is the parent Bone (-1 for root)
	// unk_1 is the child bone (-1 for none)
	// unk_2 is a sibling bone (-1 for none)
	//
	// Update : This seems to be true, i don't see a reason for this, but it SEEMS TO BE TRUE
	//
	int					mParent;							// if this is -1, the bone is the rootbone 
	int					mChild;								// if this is -1, there is no childbone
	int					mSibling;							// if this is -1, no siblings

	Vec3				mXAxis;								// i think this is the rotation matrix
	Vec3				mYAxis;
	Vec3				mZAxis;

	Vec3				mPosition;							// this should be the Position offset
};




//
// This is just a structure for faster lookup in Memory
//
struct LH3D_MemVertex
{
	Vec3				mPosition;				// Position (bodyspace)
	Vec2				mTexCoords;				// Texture Coordinates
	Vec3				mNormal;				// Normal (bodyspace)
	int					mnBone;
};



//
// This is just a structure for faster lookup in Memory
// 
struct LH3D_GeoModel
{
	int						mnTriangles;
	LH3D_Triangle*			mpTriangles;

	int						mSkinIndex;

	int						mnVertices;
	LH3D_MemVertex*			mpVertices;
	LH3D_MemVertex*			mpXformedVertices;
};



//
// This is just a structure for faster lookup in Memory
// 
struct LH3D_SubModel
{
	enum eModelFlags
	{
		FLAGS_NONE			= 0,
		FLAGS_BLEND			= 1,
		FLAGS_NODRAW		= 2
	};

	int						mnSubModels;
	LH3D_GeoModel*			mpSubModels;

	DWORD					mFlags;								// Flags

	int						mnBones;
	LH3D_Bone*				mpBones;
	LH3D_Bone*				mpXformedBones;
};



//
// This is just a structure for faster lookup in Memory
//
struct LH3D_Model
{
	int					mLoaded;							// Is the Model Loaded
															// 2 means from G3D

	char				mFilename[256];						// Filename of the model

	char*				mBuffer;							// The Buffer
	int					mSize;

	int					mnSkins;							// Number of Skins
	DWORD*				mpSkinIDs;							// Pointer to the SkinIDs
	DWORD*				mpSkinBinds;						// GL Binds for the Skins

	int					mnMeshes;
	LH3D_SubModel*		mpMeshes;
};


#endif//_L3D_H_