// Copyright(c) 2oo1 by Mental Vortex
#ifndef _MATH_H_
#define _MATH_H_
// 
// Description:
// 
// Some Basic Math
// 
//
// 15. April 2oo1 :
// Edited by Matthew T. Linehan 
//
// 
// 
// 



// useful macros
#define ERROR_EPSILON		(vec)(0.01f)
#define ERROR_EPSILON2		(vec)(0.0001f)

#define PI_CONST			(vec)( 3.141592653589793238462643383280f)
#define PI_CONST2			(vec)( 6.283185307179586476925286766559f)


#ifndef vec
#define vec float
#endif//vec


//
// A 2D Vector
//
class Vec2
{     
	friend class Vec3;
	friend class Mat4;

private:
    vec				v[2];

public:
	inline 			Vec2		( void )						{ set( 0, 0 );		 }
	inline 			Vec2		( vec x, vec y )				{ set( x, y);		 } 
        
	inline void		set			( vec x, vec y )				{ v[0] = x; v[1] = y; }
    inline void		set			( const vec *vP )				{ v[0] = vP[0]; v[1] = vP[1]; }
	
    inline vec*		raw			( void )						{ return v;    }                
    inline Vec2		operator+	( const Vec2& v2 ) const		{ return Vec2( v[0] + v2.v[0], v[1] + v2.v[1] ); }
    inline Vec2		operator-	( const Vec2& v2 ) const		{ return Vec2( v[0] - v2.v[0], v[1] - v2.v[1] ); }
    inline Vec2		operator*	( vec s ) const					{ return Vec2( v[0] * s, v[1] * s ); }
};



//
// A 3D Vector
//
class Vec3
{
	friend class Vec2;       
	friend class Mat4;

private:
    vec				v[3];

public:
					Vec3		( void )						{ set( 0, 0, 0);   }
					Vec3		( vec x, vec y, vec z )			{ set( x, y, z );  }
	inline void		set			( vec x, vec y, vec z )			{ v[0] = x; v[1] = y; v[2] = z; }
	inline void		set			( const vec *vP )				{ v[0] = vP[0]; v[1] = vP[1]; v[2] = vP[2]; }
    
    inline vec&		operator[]	( int i )						{ return v[i]; }
	inline vec*		raw			( void )						{ return v; }                
    
    inline Vec3		operator+	( const Vec3& v3 ) const		{ return Vec3( v[0] + v3.v[0] , v[1] + v3.v[1] , v[2] + v3.v[2] ); }
    inline Vec3		operator-	( const Vec3& v3 ) const		{ return Vec3( v[0] - v3.v[0] , v[1] - v3.v[1] , v[2] - v3.v[2] ); }
	inline Vec3		operator*	( vec s ) const					{ return Vec3( v[0] * s , v[1] * s , v[2] * s ); }
};




//
// A 3D Matrix
//
#define Mat4_identity()	Mat4( 1 , 0 , 0 , 0 ,	\
							  0 , 1 , 0 , 0 ,	\
							  0 , 0 , 1 , 0 ,	\
							  0 , 0 , 0 , 1 )

class Mat4
{
friend class Vec2;
friend class Vec3;

public:
	vec					m[16];

public:
	inline				Mat4			( void )			{ set( Mat4_identity() ); }


	inline void			set				( const Mat4& t )
	{
        m[0] = t.m[0]; m[4] = t.m[4]; m[8] =  t.m[ 8]; m[12] = t.m[12]; 
        m[1] = t.m[1]; m[5] = t.m[5]; m[9] =  t.m[ 9]; m[13] = t.m[13]; 
        m[2] = t.m[2]; m[6] = t.m[6]; m[10] = t.m[10]; m[14] = t.m[14]; 
        m[3] = t.m[3]; m[7] = t.m[7]; m[11] = t.m[11]; m[15] = t.m[15]; 
	}

	inline				Mat4			( const vec m0,  const vec m1,  const vec m2,  const vec m3, 
										  const vec m4,  const vec m5,  const vec m6,  const vec m7,
										  const vec m8,  const vec m9,  const vec m10, const vec m11, 
										  const vec m12, const vec m13, const vec m14, const vec m15 )
	{
        set( m0,  m1,  m2,  m3, 
			 m4,  m5,  m6,  m7, 
			 m8,  m9,  m10, m11,
			 m12, m13, m14, m15 );
	}

	inline void			set				( const vec m0,  const vec m1,  const vec m2,  const vec m3, 
										  const vec m4,  const vec m5,  const vec m6,  const vec m7,
										  const vec m8,  const vec m9,  const vec m10, const vec m11, 
										  const vec m12, const vec m13, const vec m14, const vec m15 )
	{
        m[0] =  m0; m[4] =  m1; m[8]  =  m2; m[12] =  m3; 
        m[1] =  m4; m[5] =  m5; m[9]  =  m6; m[13] =  m7; 
        m[2] =  m8; m[6] =  m9; m[10] = m10; m[14] = m11; 
        m[3] = m12; m[7] = m13; m[11] = m14; m[15] = m15; 
	}

	inline Vec3			operator*		( const Vec3 &n ) const
	{
        return Vec3( n.v[0]*m[0] + n.v[1]*m[4] + n.v[2]*m[ 8] + m[12], 
                     n.v[0]*m[1] + n.v[1]*m[5] + n.v[2]*m[ 9] + m[13], 
                     n.v[0]*m[2] + n.v[1]*m[6] + n.v[2]*m[10] + m[14] );
	}

	inline vec*			raw				( void )					{ return m; }
};

#endif // _MATH_H_